#include "headers/memory.h"

// Returns 1 if the `address` parameter is within the `start` and `start + size` bounds
// and sets the `memory_out` pointer to the corresponding offset in the `memory` buffer
int check_memory_bounds(uint32_t address, uint32_t start, uint32_t size, uint8_t* memory, uint8_t** memory_out) {
    if (address >= start && address < start + size) {
        *memory_out = &memory[address - start];
        return 1;
    }
    return 0;
}

// Returns a pointer to the buffer (memory)
// corresponding to the `address` parameter
uint8_t* get_memory_for_address(cpu_t* cpu, uint32_t address) {     
    uint8_t* memory;

    if (check_memory_bounds(address, ROM_ADDRESS, cpu->mem->rom_size, cpu->mem->rom, &memory)) return memory;
    if (check_memory_bounds(address, RAM_ADDRESS_P0, RAM_SIZE, cpu->mem->ram, &memory)) return memory;
    if (check_memory_bounds(address, RAM_ADDRESS_P1, RAM_SIZE, cpu->mem->ram, &memory)) return memory;
    if (check_memory_bounds(address, RAM_ADDRESS_P2, RAM_SIZE, cpu->mem->ram, &memory)) return memory;
    if (check_memory_bounds(address, VRAM_ADDRESS, VRAM_SIZE, cpu->disp->vram, &memory)) return memory;
    if (check_memory_bounds(address, KB_ADDRESS, KB_SIZE, cpu->keyboard->keyboard_registers, &memory)) return memory;
    if (check_memory_bounds(address, MALLOC_ADDRESS, MALLOC_SIZE, cpu->malloc_manager->mallocs, &memory)) return memory;
    if (check_memory_bounds(address, FKEYICON_ADDRESS, FKEYICON_SIZE, cpu->disp->fkey_icons, &memory)) return memory;
    if (check_memory_bounds(address, 0xFFFFFE00, 511, cpu->mem->tmp, &memory)) {
        printf("[Warning] Access to interrupt registers at 0x%08X\n", address);
        return memory;
    }

    printf("Request for memory at address 0x%08X is out of bounds\n", address);
    cpu->execution_finished = 1;
    return &cpu->mem->tmp[0];
}

// Write to the memory
void mem_write(cpu_t* cpu, uint32_t address, uint32_t data, uint8_t bytes) {
    // LCD input registers
    if (address == LCD_SELECT_REGISTER || address == LCD_DATA_REGISTER) {
        emulate_lcd_register_write(cpu, address, data);
        return;
    }
    
    uint8_t* mem = get_memory_for_address(cpu, address);
    // printf("Memory write [0x%08X]: : 0x%X\n", address, data);

    if (bytes == 1) {
        *mem = data & 0xFF;
    }
    else if (bytes == 2) {
        *(mem + 0) = (data >> 8) & 0xFF;
        *(mem + 1) = data & 0xFF;
    }
    else if (bytes == 4) {
        *(mem + 0) = (data >> 24) & 0xFF;
        *(mem + 1) = (data >> 16) & 0xFF;
        *(mem + 2) = (data >> 8) & 0xFF;
        *(mem + 3) = data & 0xFF;
    }
}

// Read from the memory
uint32_t mem_read(cpu_t* cpu, uint32_t address, uint8_t bytes) {
    // SH3 keyboard emulation
    if (address == KB_PORTA) {
        return emulate_keyboard_register_read(cpu);
    }
    // SH4 keyboard emulation
    else if (address >= KEYSC_ADDRESS && address < KEYSC_ADDRESS + KEYSC_SIZE) {
        int row = address - KEYSC_ADDRESS;
        int val;

        if (bytes == 1) 
            val = cpu->keyboard->row_state[row + (row % 2 == 0 ? 1 : -1)];
        else
            val = cpu->keyboard->row_state[row] | (cpu->keyboard->row_state[row+1] << 8);

        return val;
    }
    // Address that stores the model of the calculator (0 for emulator)
    else if (address == 0x80000300) {
        return 0;
    }
    // Address checked by C.Basic and Ftune to detect for emulator (0 for emulator)
    else if (address == 0x8000FFD0) {
        return 0;
    }
    // Address checked by C.Basic (ETMU register)
    else if (address == 0xA44D00D8) {
        printf("[Warning] Access to ETMU register at 0x%08X\n", address);
        return 0;
    }

    uint8_t* mem = get_memory_for_address(cpu, address);
    uint32_t data;

    if (bytes == 1) {
        data = *mem;
    }
    else if (bytes == 2) {
        data = (*mem << 8) | (*(mem + 1));
    }
    else {
        data = (*mem << 24) | (*(mem + 1) << 16) | (*(mem + 2) << 8) | (*(mem + 3));
    }

    // if (address != cpu->pc && address != cpu->pc + 1) printf("Memory read [0x%08X]: : 0x%X\n", address, data);

    return data;
}

// Emulate keyboard register A output by reading values
// from port B and M control registers.
// Returns the state of the keyboard for the current row
// https://www.planet-casio.com/Fr/forums/topic17509-2-emulateur-fx-9860-sh4.html#194390
uint32_t emulate_keyboard_register_read(cpu_t* cpu) {
    uint16_t PORTB_CTRL = mem_read(cpu, KB_PORTB_CTRL, 2);
    uint8_t row = 0;

    if (PORTB_CTRL != 0xAAAA) { // Rows 0-7
        uint16_t smask = PORTB_CTRL ^ 0xAAAA;

        while (smask >> (row*2) > 0b11) row++;
    }
    else { // Rows 8-9
        uint16_t PORTM_CTRL = mem_read(cpu, KB_PORTM_CTRL, 2); // row 8: xxxx0001 row 9: xxx0010
        
        row = 8 + ((PORTM_CTRL & 0b11) >> 1);
    }

    if (row > 9) {
        printf("[Warning] Request for keyboard row %d which does not exist!\n", row);
        return 0xffffffff;
    }
    else {
        // printf("Keyboard row %d: 0x%08X\n", row, cpu->keyboard->row_state[row]);
        return ~cpu->keyboard->row_state[row];
    }
}

// Emulate register selection and data write to the LCD screen
// Register 1: Set counter (Graph 35+)
// Register 4/8: Set row (Graph 35+/35+eII)
// Register 6: Set contrast (Graph 35+)
// Register 7/10: Write data (Graph 35+/35+eII)
// https://bible.planet-casio.com/common/hardware/lcd/T6K11.pdf
void emulate_lcd_register_write(cpu_t* cpu, uint32_t address, uint32_t data) {
    static int selected_register = 0;
    static int col = 0;
    static int row = 0;

    // Register selection
    if (address == LCD_SELECT_REGISTER) {
        selected_register = data;

        if (data != 1 && data != 4 && data != 6 && data != 7 && data != 8 && data != 10) {
            printf("[Warning] write to LCD_SELECT_REGISTER %d not implemented\n", data);
        }
    }
    // Data write
    else if (address == LCD_DATA_REGISTER) { 
        // Graph 35+ (4), Graph 35+ E II (8)
        if (selected_register == 4 || selected_register == 8) {
            static int counter = 1;
            if (counter) {
                // Set new row
                col = 0;
                row = data & 0x3F;
            }
            counter = !counter;
        }
        // Graph 35+ (7), Graph 35+ E II (10)
        else if (selected_register == 7 || selected_register == 10) {
            if (row * 16 + col >= VRAM_SIZE) {
                printf("[Warning] LCD_DATA_REGISTER write to invalid position col: %d, row: %d\n", col, row);
                return;
            }
            // Write row to LCD buffer
            cpu->disp->lcd[row * 16 + col++] = data;
        }
    }
}