#include "../headers/syscalls/bdisp.h"

// Set the pixel at the specified position in the LCD buffer.
void set_lcd_pixel(cpu_t* cpu, int x, int y, int pixel) {
	if (x < 0 || x >= SCREEN_WIDTH || y < 0 || y >= SCREEN_HEIGHT) return;
	uint32_t id = x/8 + y * 16;
	uint8_t bit = 7 - x % 8;
	update_bit(cpu->disp->lcd[id], bit, pixel);
}

// Clear the content of the screen
void syscall_Bdisp_AllClr_DD(cpu_t* cpu) {
	// printf("Run syscall: Bdisp_AllClr_DD\n");
    for (int i = 0; i < VRAM_SIZE; i++) {
        cpu->disp->lcd[i] = 0;
    }
}

// Clear the content of the VRAM
void syscall_Bdisp_AllClr_VRAM(cpu_t* cpu) {
	// printf("Run syscall: Bdisp_AllClr_VRAM\n");
    for (int i = 0; i < VRAM_SIZE; i++) {
        cpu->disp->vram[i] = 0;
    }
}

// Clear the content of the VRAM and the screen
void syscall_Bdisp_AllClr_DDVRAM(cpu_t* cpu) {
    syscall_Bdisp_AllClr_DD(cpu);
    syscall_Bdisp_AllClr_VRAM(cpu);
}

// Transfer the contents of VRAM to the screen.
void syscall_Bdisp_PutDisp_DD(cpu_t* cpu) {
    // printf("Run syscall: Bdisp_PutDisp_DD\n");
	for (int i = 0; i < VRAM_SIZE; i++) {
        cpu->disp->lcd[i] = cpu->disp->vram[i];
    }
}

void syscall_Bdisp_GetDisp_VRAM(cpu_t* cpu, uint32_t data_ptr) {
	// printf("Run syscall: Bdisp_GetDisp_VRAM (0x%08X)\n", data_ptr);

	uint8_t* data = get_memory_for_address(cpu, data_ptr);

	for (int i = 0; i < VRAM_SIZE; i++) {
		data[i] = cpu->disp->vram[i];
	}
}

/**
 * Set or erase a dot at the specified position of VRAM and/or DD (Display Driver).
 * @param x (0~127) This is the x coordinate of the specified position.
 * @param y (0~63) This is the y coordinate of the specified position.
 * @param point If you set point to 1 then the dot is made black. If you set point to 0 then the dot is cleared.
 */
void syscall_Bdisp_SetPoint_VRAM(cpu_t* cpu, uint32_t x, uint32_t y, uint32_t point) {
	if (x < 0 || x >= SCREEN_WIDTH || y < 0 || y >= SCREEN_HEIGHT) {
		// printf("[Warning] In syscall Bdisp_SetPoint_VRAM: Invalid coordinates: (%d, %d)\n", x, y);
		return;
	}

    uint32_t vramID = x/8 + y * 16;
    uint8_t bit = 7 - x % 8;
	update_bit(cpu->disp->vram[vramID], bit, point); 
}

/**
 * Draws a line to VRAM.
 * Code from the MonochromeLib library's ML_Line() function.
 * @param x1 (0 ~ 127) x starting point
 * @param y1 (0 ~ 63) y starting point
 * @param x2 (0 ~ 127) x ending point
 * @param y2 (0 ~ 63) y ending point
 * @param color (0 or 1) 0 = erase, 1 = draw
 */
void draw_line(cpu_t* cpu, int x1, int y1, int x2, int y2, int color) {
	int x = x1;
	int y = y1;
	int dx = x2 - x1;
	int dy = y2 - y1;
	int sx = dx < 0 ? -1 : 1;
	int sy = dy < 0 ? -1 : 1;

	dx = abs(dx);
	dy = abs(dy);
    syscall_Bdisp_SetPoint_VRAM(cpu, x, y, color);

	if(dx > dy) {
		int cumul = dx / 2;
		for(int i = 1; i <= dx; i++) {
			x += sx;
			cumul += dy;
			if(cumul > dx) {
				cumul -= dx;
				y += sy;
			}
            syscall_Bdisp_SetPoint_VRAM(cpu, x, y, color);
		}
	}
	else {
		int cumul = dy / 2;
		for(int i = 1; i <= dy; i++) {
			y += sy;
			cumul += dx;
			if(cumul > dy) {
				cumul -= dy;
				x += sx;
			}
            syscall_Bdisp_SetPoint_VRAM(cpu, x, y, color);
		}
	}
}

// Draws a black line to VRAM.
void syscall_Bdisp_DrawLine_VRAM(cpu_t* cpu, uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2) {
	// printf("Run syscall: Bdisp_DrawLine_VRAM (%d %d %d %d)\n", x1, y1, x2, y2);
	draw_line(cpu, x1, y1, x2, y2, 1);
}

// Draws a white line to VRAM.
void syscall_Bdisp_ClearLine_VRAM(cpu_t* cpu, uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2) {
	// printf("Run syscall: Bdisp_ClearLine_VRAM (%d %d %d %d)\n", x1, y1, x2, y2);
	draw_line(cpu, x1, y1, x2, y2, 0);
}

// Reverses the specified area of VRAM.
void syscall_Bdisp_AreaReverseVRAM(cpu_t* cpu, uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2) {
	// printf("Run syscall: Bdisp_AreaReverseVRAM (%d %d %d %d)\n", x1, y1, x2, y2);

	for (int y = y1; y <= y2; y++) {
		for (int x = x1; x <= x2; x++) {
			uint32_t vramID = x/8 + y * 16;
			uint8_t bit = 7 - x % 8;
			if (x < 0 || x >= SCREEN_WIDTH || y < 0 || y >= SCREEN_HEIGHT) continue;
			toggle_bit(cpu->disp->vram[vramID], bit);
		}
	}
}

/**
 * Copy a bitmap from a source rectangle to VRAM.
 * @param graph_ptr This is the pointer to the DISPGRAPH structure. See
 * https://wiki.planet-casio.com/fr/Fxlib.h#Bdisp_WriteGraph_DD.2F_Bdisp_WriteGraph_VRAM.2F_Bdisp_WriteGraph_DDVRAM
 * @remarks The bitmap data format is monochrome. The monochrome bitmap uses a one-bit, one- pixel format. Each scan is a
 * multiple of 8 bits. If the corresponding bit in the bitmap is 1, the pixel is set to black. If the corresponding bit in
 * the bitmap is zero, the pixel is set to white. Pixels are stored starting in the top left corner going from left to right
 * and then row by row from the top the bottom.
 */
void syscall_Bdisp_WriteGraph_VRAM(cpu_t* cpu, uint32_t graph_ptr) {
	uint32_t start_x = mem_read(cpu, graph_ptr, 4);
	uint32_t start_y = mem_read(cpu, graph_ptr + 4, 4);
	uint32_t width = mem_read(cpu, graph_ptr + 8, 4);
	uint32_t height = mem_read(cpu, graph_ptr + 12, 4);
	uint32_t bitmap_ptr = mem_read(cpu, graph_ptr + 16, 4);
	uint8_t  write_modify = mem_read(cpu, graph_ptr + 20, 1);
	uint8_t  write_kind = mem_read(cpu, graph_ptr + 21, 1);

	if ((write_modify != IMB_WRITEMODIFY_NORMAL && write_modify != IMB_WRITEMODIFY_REVERSE) || (write_kind != IMB_WRITEKIND_OVER && write_kind != IMB_WRITEKIND_OR)) {
		printf("[Warning] In syscall Bdisp_WriteGraph_VRAM: write_modify or write_kind not implemented (%d, %d)\n", write_modify, write_kind);
		return;
	}

	// printf("Run syscall: Bdisp_WriteGraph_VRAM (x: %d, y: %d, w: %d, h: %d, modify: %d, kind: %d)\n", start_x, start_y, width, height, write_modify, write_kind);

	uint8_t* bitmap_data = get_memory_for_address(cpu, bitmap_ptr);
	
	for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
			int id = y * (width < 8 ? 8 : width) + x;
			int packed_id = id / 8;
			int bit = 7 - id % 8;
			int val = get_bit(bitmap_data[packed_id], bit);
			if (write_modify == IMB_WRITEMODIFY_REVERSE) val = !val;

			if (write_kind == IMB_WRITEKIND_OR) {
				int vram_id = (y + start_y) * 16 + (x + start_x) / 8;
				int vram_bit = 7 - (x + start_x) % 8;
				if (vram_id < 0 || vram_id >= VRAM_SIZE) continue;

				int vram_pixel = get_bit(cpu->disp->vram[vram_id], vram_bit);
				val |= vram_pixel;
			}
			syscall_Bdisp_SetPoint_VRAM(cpu, x + start_x, y + start_y, val);
		}
	}
}

/**
 * Saves a screen image in the system area.
 * @param num SAVEDISP_PAGE1 - SAVEDISP_PAGE3.
 * @remarks The screen image is copied from VRAM.
 */
void syscall_SaveDisp(cpu_t* cpu, unsigned char num) {
	num = num == SAVEDISP_PAGE1 ? num - 1 : num - 4;

	// printf("Run syscall: SaveDisp (%d)\n", num);

	memcpy(cpu->disp->saved_disps + VRAM_SIZE*num, cpu->disp->vram, VRAM_SIZE);
}

/**
 * Restores a screen image from the system area.
 * @param num SAVEDISP_PAGE1 - SAVEDISP_PAGE3.
 * @remarks The screen image is copied to VRAM.
 */
void syscall_RestoreDisp(cpu_t* cpu, unsigned char num) {
	num = num == SAVEDISP_PAGE1 ? num - 1 : num - 4;

	// printf("Run syscall: RestoreDisp (%d)\n", num);

	memcpy(cpu->disp->vram, cpu->disp->saved_disps + VRAM_SIZE*num, VRAM_SIZE);
}

/**
 * The PopUpWin function displays a popup window of the specified size of lines. Also, the inside of the popup window is cleared.
 * @param n (1~5) This is the number of lines of the popup window.
 * @remarks To display the contents (string, etc.) of a window, use the Locate function or the Print function.
 */
void syscall_PopupWin(cpu_t* cpu, uint32_t n_lines) {
	if (n_lines < 1 || n_lines > 6) {
		printf("[Warning] In syscall PopuWin: Invalid number of lines: %d\n", n_lines);
		return;
	}

	printf("Run syscall: PopupWin (%d lines)\n", n_lines);

	// Calculate the bounds of the popup window
	const int x_min = 9;
	const int x_max = 119;
	int y_min = 19 - n_lines / 2 * 8;
	int y_max = 36 + (n_lines - 1) / 2 * 8 + 1;
	if (n_lines == 6) y_min += 8, y_max += 8;

	// Draw the popup window
	for (int y = y_min; y < y_max; y++) {
		for (int x = x_min; x < x_max; x++) {
			if (x == x_max-1 && y == y_min || x == x_min && y == y_max-1) continue;

			int white_pixel = x != x_min && x < x_max-2 && y != y_min && y < y_max-2;

			syscall_Bdisp_SetPoint_VRAM(cpu, x, y, !white_pixel);
		}
	}
}

// Displays the 4x4 busy-square (top-right). It is written directly to the DD. So the next VRAM to DD automatically clears it (f. i. GetKey).
void syscall_HourGlass(cpu_t* cpu) {
	printf("Run syscall: HourGlass\n");

	set_lcd_pixel(cpu, SCREEN_WIDTH, SCREEN_HEIGHT, 1);
	set_lcd_pixel(cpu, SCREEN_WIDTH-1, SCREEN_HEIGHT, 1);
	set_lcd_pixel(cpu, SCREEN_WIDTH, SCREEN_HEIGHT-1, 1);
	set_lcd_pixel(cpu, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, 1);
}

// Returns the pointer to the FKey-icon defined by FKeyNo in bitmap_ptr
void syscall_GetFKeyIconPointer(cpu_t* cpu, uint32_t key, uint32_t bitmap_ptr) {
	int id = key * 24;

	if (id < 0 || id >= FKEYICON_SIZE) {
		printf("[Warning] In syscall GetFKeyIconPointer: Invalid key: %d\n", key);
		id = 1016;
	}
	// else printf("Run syscall: GetFKeyIconPointer (%d, 0x%08X)\n", key, bitmap_ptr);
	
	mem_write(cpu, bitmap_ptr, FKEYICON_ADDRESS + id, 4);
}

// Displays the FKey-icon defined by pBitMap at position FKeyPos
void syscall_DisplayFKeyIcon(cpu_t* cpu, uint32_t key_pos, uint32_t bitmap_ptr) {
	// printf("Run syscall: DisplayFKeyIcon (%d, 0x%08X)\n", key_pos, bitmap_ptr);

	for (int y = 0; y < FKEYICON_HEIGHT; y++) {
		for (int x = 0; x < FKEYICON_WIDTH; x++) {
			int id = y * FKEYICON_WIDTH + x;
			int packed_id = id / 8;
			int bit = 7 - id % 8;
			int val = get_bit(mem_read(cpu, bitmap_ptr + packed_id, 1), bit);
			syscall_Bdisp_SetPoint_VRAM(cpu, x + key_pos*21 + 2, y + SCREEN_HEIGHT - FKEYICON_HEIGHT, val);
		}
	}
}