#include "../headers/syscalls/filesystem.h"
#include <dirent.h> 
#include <ctype.h>

const char* permanent_path(const char* filename) {
    char* path = malloc(sizeof(char) * (strlen(filename) + 7));
    strcpy(path, "/idbfs/");
    strcat(path, filename);
    return (const char*)path;
}

// Lowercase the file extension
const char* lower_extension(const char* str) {
    char* new_str = strdup(str);
    int ext = 0;
    for (int i = 0; i < strlen(new_str); i++) {
        if (new_str[i] == '.') ext = 1;
        else if (ext) new_str[i] = tolower(new_str[i]);
    }
    return (const char*)new_str;
}

const char* convertFileName(const char* filename, int trim) {
    char name[40];
    char ch;
    int i;

    // Convert filename from { 0x00, 0xXX, 0x00, 0xXX, ... } to { 0xXX, 0xXX, ... }
    for (i = 0; i < 40; i++) {
        uint16_t ch = *(uint16_t*)&filename[i * 2 + 1];
        name[i] = ch;
        if (ch == 0x00) break;
    }

    // Remove '\\fls0\'
    if (trim) i -= 6;
    char* final_name = malloc(sizeof(char) * i);
    memcpy(final_name, name + 7 * trim, i);

    return (const char*)final_name;
}

/**
 * Creates a file/directory
 * @param filename This is the pointer to a null-terminated string that names the file/directory to be created.
 * @param size This is the size of the file to be created. This parameter is only used when you create a file in Storage Memory.
 * @return If the function succeeds, the return value is 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks <p>This function creates a file in Storage Memory or the SD card.
 */
void syscall_Bfile_CreateEntry(cpu_t* cpu, uint32_t filename_ptr, int mode, uint32_t size_ptr) {
    const char* filename = (const char*)get_memory_for_address(cpu, filename_ptr);
    const char* name = convertFileName(filename, 1);

    if (mode == 0x1) { 
        // Create file with fixed size, initialize with 0
        uint32_t size = mem_read(cpu, size_ptr, 4);
        uint8_t* data = calloc(size, sizeof(uint8_t));
        FILE* file = fopen(permanent_path(name), "wb");
        fwrite(data, 1, size, file);
        fclose(file);

        cpu->fs->needs_sync = 1;
        R0 = 0; // Success

        printf("Run syscall: Bfile_CreateEntry %s (0x%X, size: %d)\n", name, mode, mem_read(cpu, size_ptr, 4));
    }
    // else if (mode == 0x5) { } // Create folder
    else {
        R0 = -1; // Error
        printf("[Warning] In syscall Bfile_CreateEntry: CreateEntry mode %d not implemented\n", mode);
    }
}

/**
 * Deletes an existing file
 * @param filename This is the pointer to a null-terminated string that specifies the file to delete.
 * @param mode always should be used and set to 0.
 * @return If the function succeeds, the return value is 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks <p>This function deletes a file in Storage Memory or the SD card.
 */
void syscall_Bfile_DeleteEntry(cpu_t* cpu, uint32_t filename_ptr, int mode) {
    const char* filename = (const char*)get_memory_for_address(cpu, filename_ptr);
    const char* name = convertFileName(filename, 1);

    if (remove(permanent_path(name)) == 0) {
        cpu->fs->needs_sync = 1;
        R0 = 0; // Success
        printf("Run syscall: Bfile_DeleteEntry %s\n", name);
    }
    else {
        R0 = -1; // Error
        printf("[Warning] In syscall Bfile_DeleteEntry: Could not delete file %s\n", name);
    }
}

// Used by Bfile_FindFirst and Bfile_FindNext
// Search an entry matching a search string starting from the specified index
// Writes the entry name to foundfile_ptr and the entry infos to fileinfo_ptr
int search_file(cpu_t* cpu, const char* search_string, uint32_t foundfile_ptr, uint32_t fileinfo_ptr, int index) {
    // Open root directory
    struct dirent *dir;
    DIR *d = opendir("/idbfs");

    if (d == NULL) {
        printf("[Warning] search_file: Could not open root directory\n");
        return -1;
    }

    int file_index = 0;
    int found_type = -1;
    char* found_name;

    // Lowercase the file extension
    const char* search_str = lower_extension(search_string);
    
    // Loop through each file
    while ((dir = readdir(d)) != NULL) {        
        if (((strcmp(search_str, "*.*") == 0 || strcmp(lower_extension(dir->d_name), search_str) == 0)) && file_index >= index) {
            // Matching file found
            found_type = dir->d_type;
            found_name = strdup(dir->d_name);
            break;
        }
        file_index++;
    }
    closedir(d);

    // No matching file found
    if (found_type == -1) {
        printf("[Warning] search_file: No file found for %s\n", search_str);
        return -1;
    }

    uint32_t file_size = 0;
    uint16_t file_type = DT_FILE;
    
    // Get file type
    if (strcmp(found_name, ".") == 0) file_type = DT_DOT;
    else if (strcmp(found_name, "..") == 0) file_type = DT_DOTDOT;
    else if (found_type == DT_DIR) file_type = DT_DIRECTORY;
    else if (found_type == DT_REG) file_type = DT_FILE;
    else {
        printf("[Warning] search_file: Unknown file type: %d\n", found_type);
        return -1;
    }

    if (file_type == DT_FILE) {
        // Get file size
        FILE* file = fopen(permanent_path(found_name), "rb");
        fseek(file, 0, SEEK_END);
        file_size = ftell(file);
        fclose(file);
    }

    // Write file infos
    mem_write(cpu, fileinfo_ptr, index, 2);         // (short) index
    mem_write(cpu, fileinfo_ptr + 2, file_type, 2); // (short) type (0 = folder, 1 = file)
    mem_write(cpu, fileinfo_ptr + 4, 64000, 4);     // (long)  size
    mem_write(cpu, fileinfo_ptr + 8, file_size, 4); // (long)  data size (no header)
    mem_write(cpu, fileinfo_ptr + 12, -1, 4);       // (int)   entirely filled? yes=0
    mem_write(cpu, fileinfo_ptr + 16, 0, 4);        // (long)  address

    // Write file name
    for (int i = 0; i < 20; i++) {
        mem_write(cpu, foundfile_ptr + i * 2, found_name[i], 2);
        if (found_name[i] == '\0') break;
    }

    printf("search_file: Found file %s (size: %d, type: %d)\n", found_name, file_size, file_type);

    return 0;
}

/**
 * Searches a directory for a file whose name matches the specified filename.
 * Examines subdirectory names as well as filenames.
 * @param pathname This is the pointer to a null-terminated string that specifies a valid directory or path and filename,
 * which can contain wildcard characters “*”.
 * @return If the function succeeds, the return value is 0. When a specified file is not found, this function returns
 * IML_FILEERR_ENUMRATEEND (declared in filebios.h).
 * If the function fails, the return value is an error code. It is a negative value.
 */
void syscall_Bfile_FindFirst(cpu_t* cpu, uint32_t pathname_ptr, uint32_t findhandle_ptr, uint32_t foundfile_ptr, uint32_t fileinfo_ptr) {
    const char* pathname = (const char*)get_memory_for_address(cpu, pathname_ptr);
    const char* name = convertFileName(pathname, 1);

    if (cpu->fs->search_count > 0) {
        printf("[Warning] In syscall Bfile_FindFirst: Concurrent searches not implemented (%s)\n", name);
        R0 = -1;
        return;
    }

    printf("Run syscall: Bfile_FindFirst %s\n", name);

    // Write search handle
    mem_write(cpu, findhandle_ptr, 0, 4);

    // Create search entry
    cpu->fs->searches[0].index = 0;
    cpu->fs->searches[0].handle = cpu->fs->search_count;
    strcpy(cpu->fs->searches[0].pathname, name);

    cpu->fs->search_count++;

    R0 = search_file(cpu, name, foundfile_ptr, fileinfo_ptr, 0);
}

/**
 * Continues a file search from a previous call to the Bfile_FindFirst function.
 * @param FindHandle This is a search handle returned by a previous call to the Bfile_FindFirst function.
 * @param foundfile This is the pointer to the buffer that receives the found file or subdirectory
 * @param fileinfo This is the pointer to the FILE_INFO structure that receives information about the found file or subdirectory.
 * @return If the function succeeds, the return value is 0. When a specified file is not found, this function returns
 * IML_FILEERR_ENUMRATEEND (declared in filebios.h).
 */ 
void syscall_Bfile_FindNext(cpu_t* cpu, uint32_t findhandle, uint32_t foundfile_ptr, uint32_t fileinfo_ptr) {
    if (findhandle >= cpu->fs->search_count) {
        printf("[Warning] In syscall Bfile_FindNext: Invalid search handle: 0x%X\n", findhandle);
        R0 = -1;
        return;
    }

    file_search_t* search = &cpu->fs->searches[findhandle];

    // printf("Run syscall: Bfile_FindNext #%d (%s, handle: 0x%X)\n", search->index, search->pathname, findhandle);

    R0 = search_file(cpu, search->pathname, foundfile_ptr, fileinfo_ptr, search->index + 1);
    search->index++;
}

/**
 * Closes the specified search handle.
 * @param FindHandle This is a search handle. This handle must have been previously opened by the Bfile_FindFirst function.
 * @return If the function succeeds, the return value is 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks Only four search handles can be opened at the same time.
 */
void syscall_Bfile_FindClose(cpu_t* cpu, uint32_t findhandle) {
    R0 = -1;

    uint8_t search_count = cpu->fs->search_count;
    for (int i = 0; i < search_count; i++) {
        if (R0 != 0 && cpu->fs->searches[i].handle == findhandle) {
            cpu->fs->search_count--;
            R0 = 0;
            printf("Run syscall: Bfile_FindClose (0x%0X)\n", findhandle);
        }
        else if (R0 == 0 && i < 3) {
            strcpy(cpu->fs->searches[i].pathname, cpu->fs->searches[i + 1].pathname);
            cpu->fs->searches[i].index = cpu->fs->searches[i + 1].index;
            cpu->fs->searches[i].handle = cpu->fs->searches[i + 1].handle;
            // printf("Bfile_FindClose: Moved search %s from %d to %d\n", cpu->fs->searches[i].pathname, i + 1, i);
        }
    }

    if (R0 == -1)
        printf("[Warning] In syscall Bfile_FindClose: Search not found for handle: 0x%X\n", findhandle);
}

/**
 * Opens an existing file.
 * @param filename This is the pointer to a null-terminated string that names the file to be opened.
 * @param mode The mode parameter specifies the action to open (read/write/read_write...)
 * @return If the function succeeds, the return value specifies a file handle. It is greater than or equal to 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks This function opens a file in Storage Memory or the SD card.
 * mode2 always should be used and set to 0.
 */
void syscall_Bfile_OpenFile(cpu_t* cpu, uint32_t filename_ptr, int mode, int mode2) {
    const char* filename = (const char*)get_memory_for_address(cpu, filename_ptr);
    const char* name = convertFileName(filename, 1);
    fs_t* fs = cpu->fs;

    // Only four files can be opened at the same time    
    if (fs->open_files_count >= 4) {
        printf("[Warning] In syscall Bfile_OpenFile: Too many files opened! (%s, mode: %d)\n", name, mode);
        R0 = -8;
        return;
    }

    // Check if file exists
    FILE* file = fopen(permanent_path(name), "rb");

    if (file == NULL) {
        printf("[Warning] In syscall Bfile_OpenFile: File not found: %s\n", name);
        R0 = -1;
        return;
    }

    // Get file size
    fseek(file, 0, SEEK_END);
    int32_t file_size = ftell(file);
    fclose(file);

    // Open file for reading/writing
    switch (mode) {
        case OPENMODE_READ: file = fopen(permanent_path(name), "rb"); break;
        case OPENMODE_READ_SHARE: file = fopen(permanent_path(name), "rb"); break;
        case OPENMODE_WRITE: file = fopen(permanent_path(name), "wb"); break;
        case OPENMODE_READWRITE: file = fopen(permanent_path(name), "rb+"); break;
        case OPENMODE_READWRITE_SHARE: file = fopen(permanent_path(name), "rb+"); break;
        default:
            printf("[Warning] In syscall Bfile_OpenFile: Unknown mode: %d\n", mode);
            R0 = -1;
            return;
    }

    // Handle is 0x1000000 + file index
    int32_t handle = 0x01000000 + fs->open_files_count;

    // Create file entry
    strcpy(fs->files[fs->open_files_count].name, name);
    cpu->fs->files[fs->open_files_count].handle = handle;
    cpu->fs->files[fs->open_files_count].file_handle = file;
    cpu->fs->files[fs->open_files_count].seek_pos = 0;
    cpu->fs->files[fs->open_files_count].size = file_size;
    fs->open_files_count++;
    
    printf("Run syscall: Bfile_OpenFile: %s (mode: %d), handle: 0x%X\n", name, mode, handle);

    R0 = handle;
}

/**
 * Closes an open file handle.
 * @param HANDLE This is the handle of the file to close. HANDLE should be the handle opened by the Bfile_OpenFile or
 * Bfile_OpenMainMemory function.
 * @return If the function succeeds, this function returns 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks Only four files can be opened at the same time.
 */
void syscall_Bfile_CloseFile(cpu_t* cpu, int handle) {
    R0 = -1;

    // Check if the handle is valid
    if (handle - 0x01000000 < 0 || handle - 0x01000000 >= 4) {
        printf("[Warning] Bfile_CloseFile: Invalid handle: 0x%X\n", handle);
        return;
    }

    // Loop through each open file
    uint8_t open_count = cpu->fs->open_files_count;
    for (int i = 0; i < open_count; i++) {
        // Find the corresponding file and close it
        if (R0 != 0 && cpu->fs->files[i].handle == handle) {
            fclose(cpu->fs->files[i].file_handle);
            cpu->fs->open_files_count--;
            R0 = 0;

            printf("Run syscall: Bfile_CloseFile %s (0x%X)\n", cpu->fs->files[i].name, handle);
        }
        // Shift the following files to the left
        else if (R0 == 0 && i < 3) {
            strcpy(cpu->fs->files[i].name, cpu->fs->files[i + 1].name);
            cpu->fs->files[i].handle = cpu->fs->files[i + 1].handle;
            cpu->fs->files[i].file_handle = cpu->fs->files[i + 1].file_handle;
            cpu->fs->files[i].seek_pos = cpu->fs->files[i + 1].seek_pos;
            cpu->fs->files[i].size = cpu->fs->files[i + 1].size;
            // printf("Bfile_CloseFile: Moved file %s from %d to %d\n", cpu->fs->files[i].name, i + 1, i);
        }
    }

    if (R0 == -1)
        printf("[Warning] In syscall Bfile_CloseFile: File not found for handle: 0x%X\n", handle);
}

/**
 * Writes data to a file. The function starts writing data to the file at the position
 * indicated by the file pointer. After the write operation has been completed, the file pointer is adjusted by the
 * number of bytes actually written (TODO)
 * @param HANDLE This is the handle of the file to write. HANDLE should be the handle opened by the Bfile_OpenFile or
 * Bfile_OpenMainMemory function.
 * @param buf This is the pointer to the buffer containing the data to be written to the file.
 * @param size This is the number of bytes to write to the file.
 * @return If the function succeeds, this function returns the position indicated by the file pointer.
 * It is greater than or equal to 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks If you use a file that is written by this function on the PC, you should save the multi byte data in the Little Endian
 * format.
 */
void syscall_Bfile_WriteFile(cpu_t* cpu, int handle, uint32_t buf_ptr, int size) {
    // Check if the handle is valid
    if (handle - 0x01000000 < 0 || handle - 0x01000000 >= 4) {
        printf("[Warning] In syscall Bfile_WriteFile: Invalid handle: 0x%X (0x%X, %d)\n", handle, buf_ptr, size);
        R0 = -1;
        return;
    }

    file_t* file = &cpu->fs->files[handle & 3];

    printf("Run syscall: Bfile_WriteFile %s (0x%X, 0x%X, %d)\n", file->name, handle, buf_ptr, size);

    // Advance the file pointer
    fseek(file->file_handle, file->seek_pos, SEEK_SET);
    file->seek_pos += size;

    // Write the data to the file
    for (int i = 0; i < size; i++) {
        uint8_t byte = mem_read(cpu, buf_ptr + i, 1);
        fwrite(&byte, 1, 1, file->file_handle);
    }

    cpu->fs->needs_sync = 1;
    R0 = file->seek_pos;
}

/**
 * Reads data from a file, starting at the position indicated by the file pointer. After the
 * read operation has been completed, the file pointer is adjusted by the number of bytes actually read.
 * @param HANDLE This is the handle of the file to read. HANDLE should be the handle opened by the Bfile_OpenFile or
 * Bfile_OpenMainMemory function.
 * @param buf This is the pointer to the buffer that receives the data read from the file.
 * @param size This is the number of bytes to be read from the file.
 * @param readpos This is the starting position to read. If the readpos parameter is -1, this function reads data from the position
 * indicated by the file pointer. If the readpos parameter greater than or equal to 0, this function reads data from
 * the position indicated by the readpos parameter.
 * @return If the function succeeds, this function returns the number of bytes actually read. It is greater than or equal to 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks If you read the Windows file, the multi byte data is stored in Little Endian format. To use this data, you should
 * convert the multi byte data into the Big Endian format.
 */
void syscall_Bfile_ReadFile(cpu_t* cpu, int handle, uint32_t buf_ptr, int size, int readpos) {
    // Check if the handle is valid
    if (handle - 0x01000000 < 0 || handle - 0x01000000 >= 4) {
        printf("[Warning] In stscall Bfile_ReadFile: Invalid handle: 0x%X (0x%X, %d, %d)\n", handle, buf_ptr, size, readpos);
        return;
    }

    file_t* file = &cpu->fs->files[handle & 3];

    printf("Run syscall: Bfile_ReadFile %s (0x%X, 0x%X, %d, %d)\n", file->name, handle, buf_ptr, size, readpos);

    // Advance the file pointer
    if (readpos == -1) {
        fseek(file->file_handle, file->seek_pos, SEEK_SET);
        file->seek_pos += size;
    }
    else {
        fseek(file->file_handle, readpos, SEEK_SET);
        file->seek_pos = readpos + size;
    }

    // Read the data from the file
    for (int i = 0; i < size; i++) {
        uint8_t byte;
        fread(&byte, 1, 1, file->file_handle);
        mem_write(cpu, buf_ptr + i, byte, 1);
    }

    R0 = size;
}

/**
 * Moves the file pointer of an open file.
 * @param HANDLE This is the file handle whose file pointer will be moved
 * @param pos This is the number of bytes to move file pointer.
 * @return If the function succeeds, this function returns the number of bytes that can continuously be read.
 * It is greater than or equal to 0.
 * If the function fails, the return value is an error code. It is a negative value.
 * @remarks The starting point for the file pointer move is the beginning of the file. The file pointer is at the head of the file
 * immediately after the file was opened.
 */
void syscall_Bfile_SeekFile(cpu_t* cpu, int handle, int pos) {
    // Check if the handle is valid
    if (handle - 0x01000000 < 0 || handle - 0x01000000 >= 4) {
        printf("[Warning] In syscall Bfile_SeekFile: Invalid handle: 0x%X (%d)\n", handle, pos);
        return;
    }

    file_t* file = &cpu->fs->files[handle & 3];

    // Advance the file pointer
    file->seek_pos = pos;

    // Return the number of bytes left from current pos
    R0 = file->size - pos;

    printf("Run syscall: Bfile_SeekFile %s (0x%X, %d) ret: %d\n", file->name, handle, pos, R0);
}

/**
 * Retrieves the size, in bytes, of the specified file.
 * @param handle This is the handle of the file whose size will be retrieved. HANDLE should be the handle opened by the
 * Bfile_OpenFile or Bfile_OpenMainMemory function.
 * @return If the function succeeds, the return value is the file size.
 * If the function fails, the return value is an error code. It is a negative value.
 */
void syscall_Bfile_GetFileSize(cpu_t* cpu, uint32_t handle) {
    // Check if the handle is valid
    if (handle - 0x01000000 < 0 || handle - 0x01000000 >= 4) {
        printf("[Warning] In syscall Bfile_GetFileSize: Invalid handle: 0x%X\n", handle);
        R0 = -1;
        return;
    }

    file_t* file = &cpu->fs->files[handle & 3];

    printf("Run syscall: Bfile_GetFileSize %s (0x%X) => %d\n", file->name, handle, file->size);

    R0 = file->size;
}

/**
 * Retrieves the size of the free space, in bytes, of the specified device.
 * @param media_id is either {'\\','\\','f','l','s','0',0} or {'\\','\\','c','r','d','0',0}.
 * @param freebytes This is the pointer to the size of the free space.
 * @return -5, if media_id cannot be recognized, else returns 0.
 * @remarks cannot be used to read the main memory freespace!
 * With the GII-2 models freespace is an two element int-array: int freespace[2] (int64). 
 * The free space is returned in freespace[1].
 */
void syscall_Bfile_Bfile_GetMediaFree(cpu_t* cpu, uint32_t media_id_ptr, uint32_t freespace_ptr) {
    const char* media_name = (const char*)get_memory_for_address(cpu, media_id_ptr);
    char* media_id = (char*)convertFileName(media_name, 0);
    media_id[7] = '\0';

    if (strcmp(media_id, "\\\\fls0\\") == 0) {
        printf("Run syscall: Bfile_GetMediaFree (%s) => %d\n", media_id, ROM_SIZE);
        mem_write(cpu, freespace_ptr, 0, 4);
        mem_write(cpu, freespace_ptr + 4, ROM_SIZE, 4);
        R0 = 0;
    }
    else if (strcmp(media_id, "\\\\crd0\\") == 0) {
        printf("[Warning] In syscall Bfile_Bfile_GetMediaFree: SD card (crd0) not supported\n");
        R0 = -1;
    }
    else {
        printf("[Warning] In syscall Bfile_Bfile_GetMediaFree: Unknown media: %s\n", media_id);
        R0 = -5;
    }
}

// format to "{dir}.{item}"
const char* getFilePath(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr) {
    const char* dir_name = (const char*)get_memory_for_address(cpu, dir_ptr);
    const char* item_name = (const char*)get_memory_for_address(cpu, item_ptr);
    char* file_path = malloc(sizeof(char) * (strlen(dir_name) + strlen(item_name) + 2));

    strcpy(file_path, dir_name);
    strcat(file_path, ".");
    strcat(file_path, item_name);

    return permanent_path(file_path);
}

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_BfileMainMemory.HTM
// dir directory name (f. i. "system" in case of BASIC programs)
// item item name (f. i. the name of a BASIC program)
// *data_len length of item-data
// If successful, the function sets the MCS-handles to dir and item.
// Returns 0, if no error occurs.
// Returns 0x40, if item does not exist.
void syscall_MCSGetDlen2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t len_ptr) {
    const char* file_path = getFilePath(cpu, dir_ptr, item_ptr);

    FILE* file = fopen(file_path, "rb");

    if (file == NULL) {
        printf("[Warning] In syscall MCSGetDlen2: File does not exist: %s\n", file_path);
        R0 = 0x40;
        return;
    }

    // Get file size
    fseek(file, 0, SEEK_END);
    int32_t file_size = ftell(file);
    fclose(file);
    
    mem_write(cpu, len_ptr, file_size, 4);

    cpu->fs->mcs_filename = strdup(file_path);
    cpu->fs->mcs_filesize = file_size;
	
	printf("Run syscall: MCSGetDlen2 (%s) => %d. Set MCS handle\n", file_path, file_size);
	
	R0 = 0;
}

// Operates on MCS handles.
// if offset is zero, the complete available data-block is copied to *buffer (len_to_copy is ignored!).
// Otherwise the len_to_copy bytes of data beginning at offset are copied to *buffer.
// Returns 0, if no error occurred.
void syscall_MCSGetData1(cpu_t* cpu, int32_t offset, int32_t len_to_copy, int32_t buffer_ptr) {
    if (cpu->fs->mcs_filename == NULL) {
        printf("[Warning] In syscall MCSGetData1: No MCS set\n");
        R0 = -1;
        return;
    }

    const char* file_name = cpu->fs->mcs_filename;
    uint32_t file_size = cpu->fs->mcs_filesize;

    printf("Run syscall: MCSGetData1, MCS: %s (offset: %d, len: %d)\n", file_name, offset, len_to_copy);

    FILE* file = fopen(file_name, "rb");

    if (file == NULL) {
        printf("[Warning] In syscall MCSGetData1: File does not exist: %s\n", file_name);
        R0 = 0x40;
        return;
    }

    fseek(file, offset, SEEK_SET);

    for (int i = 0; i < len_to_copy; i++) {
        uint8_t byte;
        fread(&byte, 1, 1, file);
        mem_write(cpu, buffer_ptr + i, byte, 1);
    }

    R0 = 0;
}

// Creates an item dir.item of length data_len in the MCS.
// item is initialized with *buffer.
// The function does not overwrite an existing dir.item.
// dir.item-member flags[0] is set to 0xD4.
// If the function succeeds, MCS-handles are set.
// Returns 0, if no error occurs.
// Returns 0x25, if dir.item already exists.
// Returns error 0x40, if dir does not exist.
void syscall_MCSPutVar2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t data_len, int32_t buffer_ptr) {
    const char* file_path = getFilePath(cpu, dir_ptr, item_ptr);
    
    // Check if file already exists
    FILE* file = fopen(file_path, "rb");

    if (file != NULL) {
        printf("[Warning] In syscall MCSPutVar2: File already exists: %s\n", file_path);
        R0 = 0x25;
        return;
    }

    file = fopen(file_path, "wb");

    // Check for file creation success
    if (file == NULL) {
        printf("[Warning] In syscall MCSPutVar2: Could not create file: %s\n", file_path);
        R0 = -1;
        return;
    }

    printf("Run syscall: MCSPutVar2, created new file %s (%d)\n", file_path, data_len);

    // Write data to file
    uint8_t* data = get_memory_for_address(cpu, buffer_ptr);
    fwrite(data, 1, data_len, file);
    fclose(file);

    cpu->fs->needs_sync = 1;
    R0 = 0;
}

// Writes bytes_to_write bytes of buffer to dir.item.
// write_offset defines where to start the write inside dir.item's data area.
// If the function succeeds, MCS-handles are set.
// Returns 0, if no error occurred.
void syscall_MCSOvwDat2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t bytes_to_write, int32_t buffer_ptr, int32_t write_offset) {
    const char* file_path = getFilePath(cpu, dir_ptr, item_ptr);

    // Check if file exists
    FILE* file = fopen(file_path, "rb");

    if (file == NULL) {
        printf("[Warning] In syscall MCSOvwDat2: File does not exist: %s\n", file_path);
        R0 = 0x40;
        return;
    }
    fclose(file);

    // Open file for writing
    file = fopen(file_path, "wb");

    // Check for file creation success
    if (file == NULL) {
        printf("[Warning] In syscall MCSOvwDat2: Could not create file: %s\n", file_path);
        R0 = -1;
        return;
    }

    printf("Run syscall: MCSOvwDat2, writing %d bytes to file %s (offset: %d)\n", bytes_to_write, file_path, write_offset);

    // Write data to file
    uint8_t* data = get_memory_for_address(cpu, buffer_ptr);
    fwrite(data + write_offset, 1, bytes_to_write, file);
    fclose(file);

    cpu->fs->needs_sync = 1;
    R0 = 0;
}

void syscall_MCSDelVar2(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr) {
    const char* file_path = getFilePath(cpu, dir_ptr, item_ptr);

    // Check if file exists
    FILE* file = fopen(file_path, "rb");

    if (file == NULL) {
        printf("[Warning] In syscall MCSDelVar2: File does not exist: %s\n", file_path);
        R0 = 0x40;
        return;
    }
    fclose(file);

    // Delete file
    if (remove(file_path) == 0) {
        cpu->fs->mcs_filename = NULL;
        cpu->fs->needs_sync = 1;
        printf("Run syscall: MCSDelVar2, deleted file %s. MCS cleared\n", file_path);
        R0 = 0;
    }
    else {
        printf("[Warning] In syscall MCSDelVar2: Could not delete file: %s\n", file_path);
        R0 = -1;
    }
}

// Returns max. available main memory in *maxspace.
// Returns current main memory load in *currentload.
// Returns currently availabe main memory in *remainingspace.
// Returns 0.
void syscall_MCS_GetState(cpu_t* cpu, int32_t maxspace_ptr, int32_t currentload_ptr, int32_t remainingspace_ptr) {
    const int maxspace = 65536;
    const int currentload = 0;
    const int remainingspace = maxspace - currentload;

    mem_write(cpu, maxspace_ptr, maxspace, 4);
    mem_write(cpu, currentload_ptr, currentload, 4);
    mem_write(cpu, remainingspace_ptr, remainingspace, 4);

    printf("Run syscall: MCS_GetState (%d, %d, %d)\n", maxspace, currentload, remainingspace);

    R0 = 0;
}

// Tries to locate item in the MCS directory dir and if successful, retrieves some information of item.
// Returns the pointer to the item in *item_ptr.
// Returns the pointer to the data in *data_ptr.
// Returns the length of the data in *data_length.
// Returns 0, if no error occurs.
// Returns 0x30, if item not found.
void syscall_MCS_SearchDirectoryItem(cpu_t* cpu, int32_t dir_ptr, int32_t item_name_ptr, int32_t flags_ptr, int32_t item_ptr, int32_t data_ptr, int32_t len_ptr) {
    const char* item_path = getFilePath(cpu, dir_ptr, item_name_ptr);

    // Check if file exists
    FILE* file = fopen(item_path, "rb");

    if (file == NULL) {
        printf("[Warning] In syscall MCS_SearchDirectoryItem: File does not exist: %s\n", item_path);
        mem_write(cpu, len_ptr, 0, 4);
        R0 = 0x30;
        return;
    }

    // Get file size
    fseek(file, 0, SEEK_END);
    int32_t file_size = ftell(file);
    fseek(file, 0, SEEK_SET);

    printf("Run syscall: MCS_SearchDirectoryItem %s (0x%8X, 0x%8X, 0x%8X, 0x%8X) => %d\n", item_path, flags_ptr, item_ptr, data_ptr, len_ptr, file_size);

    // TODO: Improve this (quick hack)
    static int mcs_storage = 0;
    if (mcs_storage == 0) {
        syscall_Malloc(cpu, 32 * 1024, 0);
        mcs_storage = R0;
    }

    for (int i = 0; i < file_size; i++) {
        uint8_t byte;
        fread(&byte, 1, 1, file);
        mem_write(cpu, mcs_storage + i, byte, 1);
    }

    fclose(file);

    mem_write(cpu, data_ptr, mcs_storage, 4);
    mem_write(cpu, item_ptr, 0x000000bb, 4); // ???
    mem_write(cpu, len_ptr, file_size, 4);

    R0 = 0;
}

// Writes bytes_to_write bytes of buffer to dir.item.
// write_offset defines where to start the write inside dir.item's data area.
// returns 0x13, when the number of bytes to write exceeds the item's actual data length.
// returns 0x14, when the bytes_to_write plus write_offset exceeds the item's actual data block end.
// returns 0, if no error occurred.
void syscall_MCS_OverwriteData(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr, int32_t write_offset, int32_t bytes_to_write, int32_t buffer_ptr) {
    const char* item_path = getFilePath(cpu, dir_ptr, item_ptr);

    printf("[Warning]: Run syscall MCS_OverwriteData %s (offset: %d, bytes: %d, buffer: 0x%8X)\n", item_path, write_offset, bytes_to_write, buffer_ptr);

    FILE* file = fopen(item_path, "rb");

    if (file == NULL) {
        printf("[Warning] In syscall MCS_OverwriteData: File does not exist: %s\n", item_path);
        R0 = 0x40;
        return;
    }

    // Get file size
    fseek(file, 0, SEEK_END);
    int32_t file_size = ftell(file);
    fclose(file);
    
    if (write_offset + bytes_to_write > file_size) {
        printf("[Warning] In syscall MCS_OverwriteData: Write offset + bytes to write exceeds file size: %s\n", item_path);
        R0 = bytes_to_write > file_size ? 0x13 : 0x14;
        return;
    }

    // Open file for writing
    file = fopen(item_path, "wb");
    fseek(file, write_offset, SEEK_SET);

    for (int i = 0; i < bytes_to_write; i++) {
        uint8_t byte = mem_read(cpu, buffer_ptr + i, 1);
        fwrite(&byte, 1, 1, file);
    }
    fclose(file);

    // fwrite(get_memory_for_address(cpu, buffer_ptr) + write_offset, 1, bytes_to_write, file);
    
    cpu->fs->needs_sync = 1;
    R0 = 0;
}

// Deletes dir.item.
// Returns 0, if successful.
// Returns 0xF0, if either dir or item==0 or the length of *dir or *item ==0.
// Returns 0x40, if dir not found.
// Returns 0x30, if item not found.
// Returns 0x32, if item's flags[1] bit 0 is 1.
// Returns 0x34, if item's flags[1] bit 1 or 2 is 1.
void syscall_MCS_DeleteItem(cpu_t* cpu, int32_t dir_ptr, int32_t item_ptr) {
    const char* item_path = getFilePath(cpu, dir_ptr, item_ptr);

    printf("[Warning] Skipped syscall: MCS_DeleteItem %s\n", item_path);

    R0 = 0x0;
}