#include "../headers/syscalls/keyboard.h"

void manage_keyboard_shift_alpha(cpu_t* cpu, int key) {
    if (key == KEY_ALPHA) {
        if (cpu->keyboard->alpha_mode) {
            // If alpha mode is already enabled, disable it
            cpu->keyboard->alpha_mode = 0;
        } else {
            if (cpu->keyboard->shift_mode) {
                // If shift mode is enabled, enable alpha mode permanently
                cpu->keyboard->shift_mode = 0;
                cpu->keyboard->alpha_mode = 2; // Permanent mode
            }
            else {
                // Otherwise enable alpha mode for the next char only
                cpu->keyboard->alpha_mode = 1; // Next char only
            }
        }
    }
    // Disable alpha mode if it is enabled (but not permanent) or if a special key is pressed
    else if (cpu->keyboard->alpha_mode == 1 || key == KEY_EXIT || key == KEY_EXE || key == KEY_SHIFT || key == KEY_AC) {
        cpu->keyboard->alpha_mode = 0;
    }

    if (key == KEY_SHIFT) {
        // Press shift: Toggle shift mode
        cpu->keyboard->shift_mode = !cpu->keyboard->shift_mode; 
    }
    else if (cpu->keyboard->shift_mode) {
        // Otherwise: Disable shift mode if it is enabled
        cpu->keyboard->shift_mode = 0;
    }
    
    // Update cursor style
    if (cpu->keyboard->shift_mode) cpu->disp->cursor.flash_style = 1; 
    else if (cpu->keyboard->alpha_mode) cpu->disp->cursor.flash_style = 3; 
    else cpu->disp->cursor.flash_style = 0;
}

#ifdef USE_EMSCRIPT
// Async javascript promise that returns the next pressed key.
EM_ASYNC_JS(int*, async_browser_getkey, (), {
    return await new Promise((resolve, reject) => {
        // Used to resolve the promise when loading a new .g1a file (blocking otherwise)
        window.getkeyEventResolve = resolve; 

        const try_resolve_key = (key) => {
            if (key != null) {
                // Valid key pressed, remove the event listeners and return the key
                window.removeEventListener("keydown", onKeyboardEvent);
                document.getElementById("keyboard").removeEventListener("pointerdown", onClickEvent);
                window.getkeyEventResolve = null;
                resolve(key);
            }
        };

        const onClickEvent = (event) => {
            const res = window.handleKeyboardEvent(event, "getkey");
            try_resolve_key(res);
        };
        
        const onKeyboardEvent = (event) => {
            const res = window.onKeyUpdate(event, "getkey");
            try_resolve_key(res);
        };

        // Add a click event listener to the keyboard
        window.addEventListener("keydown", onKeyboardEvent);
        document.getElementById("keyboard").addEventListener("pointerdown", onClickEvent);
    });
})
#endif

// Gets the next pressed key into *keycode.
// returns 0 in case of a CTRL-key.
// returns 1 in case of a CHAR-key.
void syscall_GetKey(cpu_t* cpu, unsigned int keycode_address) {
    // Update display
    syscall_Bdisp_PutDisp_DD(cpu);

    // Draw cursor
    cpu->disp->cursor.blink_state = 0;
    draw_cursor(cpu); 

    int key;

    // First check if there are any keys in the keyboard buffer (set with PutKey syscall)
    if (cpu->keyboard->buffered_keys_count > 0) {
        printf("Run syscall: GetKey, used keybuffer: 0x%04X (%d)\n", cpu->keyboard->key_buffer[0], cpu->keyboard->key_buffer[0]);
        key = cpu->keyboard->key_buffer[0];

        for (int i = 0; i < cpu->keyboard->buffered_keys_count - 1; i++) {
            cpu->keyboard->key_buffer[i] = cpu->keyboard->key_buffer[i + 1];
        }
        cpu->keyboard->buffered_keys_count--;
    } 
    // Otherwise wait for user input
    else {
        printf("Run syscall: GetKey, waiting for key\n");
        int* keycode = async_browser_getkey();

        if (cpu->keyboard->shift_mode) key = keycode[1];
        else if (cpu->keyboard->alpha_mode) key = keycode[2];
        else key = keycode[0];

        printf("Return key: %d (shift: %d, alpha: %d)\n", key, cpu->keyboard->shift_mode, cpu->keyboard->alpha_mode);
    }

    // Manage shift and alpha states
    manage_keyboard_shift_alpha(cpu, key);

    // Return the keycode
    mem_write(cpu, keycode_address, key, 4);
    R0 = key;
}

// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_keyboard.htm
void syscall_KeyBoard_GetKeyWait(cpu_t* cpu, uint32_t column_ptr, uint32_t row_ptr, uint32_t type_of_waiting, uint32_t timeout_period, uint32_t menu, uint32_t keycode_ptr) {
    printf("[Warning]: Skipped syscall KeyBoard_GetKeyWait (wait type: %d, timeout: %d, menu: %d)X\n", type_of_waiting, timeout_period, menu);

    syscall_Bdisp_PutDisp_DD(cpu);
    mem_write(cpu, keycode_ptr, 0, 2);

    R0 = 0;
}

// Puts a keycode into the keybuffer.
// https://bible.planet-casio.com/simlo/chm/v20/fx_legacy_keyboard.htm
// keycode: value to inject exactly as returned by GetKey.
// mode controls the matrixcode, which are set.
// returns 0 if the function fails (keybuffer full), 1 if the function succeeded.
// - puts -1,-1 (if mode==1) or -2,-2 (if mode!=1) into the matrixcode buffer .
// - puts the GetKey-compatible keycode itself into an area beginning at 0x880061DC.
// -1,-1 and -2,-2 are no real matrix codes.
// Any special key processing like MENU, SHIFT AC/on a. s. o. is processed matrixcode-based inside of 0x247, which is called by GetKey.
// To invoke those special behaviour, matrix codes have to be injected (syscalls 0x248, 0x24F).
void syscall_PutKey(cpu_t* cpu, uint32_t keycode, uint32_t mode) {
    printf("Run syscall: PutKey #%d (0x%X (%d), mode: %d)\n", cpu->keyboard->buffered_keys_count + 1, keycode, keycode, mode);

    cpu->keyboard->key_buffer[cpu->keyboard->buffered_keys_count] = keycode;
    cpu->keyboard->buffered_keys_count++;

    R0 = 1;
}

// Clears the keyboard buffer
void syscall_Keyboard_ClrBuffer(cpu_t* cpu) {
    printf("Run syscall: Keyboard_ClrBuffer, removed %d keys\n", cpu->keyboard->buffered_keys_count);

    cpu->keyboard->buffered_keys_count = 0;
}

/**
 * Sets the repetition interval of the key.
 * The parameters are in milliseconds divided by 25.
 * @param FirstCount This is time from pushing the key to the input of the first repetition key.
 * @param NextCount This is the interval time of the repetition key.
 * @remarks The cursor key is the only key that can be set for repetition. Other keys will not repeat even if they are
 * continuously pressed.
 */
void syscall_Bkey_Set_RepeatTime(cpu_t* cpu, int32_t first_count, int32_t next_count) {
    // printf("Run syscall: Bkey_Set_RepeatTime (%d %d)\n", first_count, next_count);
    cpu->keyboard->repeat_time_first = first_count;
    cpu->keyboard->repeat_time_next = next_count;
}

/**
 * Gets the repetition interval of the key.
 * The parameters are in milliseconds divided by 25.
 * @param FirstCount This is the pointer to time from pushing the key to the input of the first repetition key.
 * @param NextCount This is the pointer to the interval time of the repetition key.
 */
void syscall_Bkey_Get_RepeatTime(cpu_t* cpu, int32_t first_count_ptr, int32_t next_count_ptr) {
    // printf("Run syscall: Bkey_Get_RepeatTime (%d %d)\n", cpu->keyboard->repeat_time_first, cpu->keyboard->repeat_time_next);
    mem_write(cpu, first_count_ptr, cpu->keyboard->repeat_time_first, 4);
    mem_write(cpu, next_count_ptr, cpu->keyboard->repeat_time_next, 4);
}

/**
 * Sets the default key repetition interval.
 * @remarks This function sets the time from pushing the key to starting key repeats to 500 milliseconds (20), 
 * and sets the interval time of repetition key to 125 milliseconds (5).
 */
void syscall_Bkey_Set_RepeatTime_Default(cpu_t* cpu) {
    // printf("Run syscall: Bkey_Set_RepeatTime_Default (%d %d)\n", DEFAULT_REPEAT_TIME_FIRST_COUNT, DEFAULT_REPEAT_TIME_NEXT_COUNT);
    cpu->keyboard->repeat_time_first = DEFAULT_REPEAT_TIME_FIRST_COUNT;
    cpu->keyboard->repeat_time_next = DEFAULT_REPEAT_TIME_NEXT_COUNT;
}