#include "../headers/syscalls/text.h"

// Detect if the current character is the first byte of a 2-byte character
int is_multibyte_char(const unsigned char c) {
    return c == 0xE5 || c == 0xE6 || c == 0x7F;
}

// Get the width of a mini character (stored as the last 3 bits of the character data)
uint8_t get_mini_char_size(uint8_t* char_data) {
    return get_bit(char_data[6], 5) | (get_bit(char_data[6], 6) << 1) | (get_bit(char_data[6], 7) << 2);
}

// Draw a character on screen
// pixel_start_x and y represent the top left pixel of the character
void draw_character(cpu_t* cpu, const unsigned char* c, int pixel_start_x, int pixel_start_y, int mode) {
    uint8_t* char_data = cpu->disp->character_set;

    if (mode < 0 || mode > 1) {
        printf("[Warning] draw_character: Invalid mode: %d\n", mode);
        mode = 0;
    }

    // Calculate offset in the character set for 2-byte characters
    if (is_multibyte_char(*c)) {
        if (*c == 0xE5) char_data += CHARACTER_SET_SIZE / 4;
        else if (*c == 0xE6) char_data += 2 * CHARACTER_SET_SIZE / 4;
        else if (*c == 0x7F) char_data += 3 * CHARACTER_SET_SIZE / 4;
        char_data += c[1] * 7;
    }
    else {
        char_data += c[0] * 7;
    }

    for (int y = 0; y < CHAR_HEIGHT; y++) {
        for (int x = 0; x < CHAR_WIDTH; x++) {
            // Current pixel position on-screen
            int vram_x = pixel_start_x + x;
            int vram_y = pixel_start_y + y;
            if (vram_x < 0 || vram_x >= SCREEN_WIDTH || vram_y < 0 || vram_y >= SCREEN_HEIGHT) break;
            
            // Get pixel state in VRAM
            int vram_id = (vram_x + vram_y * SCREEN_WIDTH) / 8; // index in VRAM
            int vram_bit = 7 - vram_x % 8;                      // Current bit
            int vram_pixel = get_bit(cpu->disp->vram[vram_id], vram_bit);

            // Get pixel state in character
            int char_pixel = (x == 0 || y == CHAR_HEIGHT - 1) ? 0 : get_bit(char_data[y], x - 1);

            // Change one single bit in the VRAM
            if (mode == TEXT_REVERSE) char_pixel = !char_pixel;
            update_bit(cpu->disp->vram[vram_id], vram_bit, char_pixel);
        }
    }
}

void draw_character_mini(cpu_t* cpu, int pixel_start_x, int pixel_start_y, uint8_t* char_data, int mode) {
    uint8_t size = get_mini_char_size(char_data);

    mode |= 0x10; // Values in 0-3 range seem to also work on the real device

    if (mode < MINI_OVER || mode > MINI_REVOR) {
        printf("[Warning] draw_character_mini: Invalid mode: %d\n", mode);
        mode = 0;
    }

    for (int y = 0; y < 6; y++) {
        for (int x = 0; x < size; x++) {
            // Current pixel position on-screen
            int vram_x = pixel_start_x + x;
            int vram_y = pixel_start_y + y;
            if (vram_x < 0 || vram_x >= SCREEN_WIDTH || vram_y < 0 || vram_y >= SCREEN_HEIGHT) break;

            // Get pixel state in VRAM
            int vram_id = (vram_x + vram_y * SCREEN_WIDTH) / 8; // index in VRAM
            int vram_bit = 7 - vram_x % 8; // Current bit
            int vram_pixel = get_bit(cpu->disp->vram[vram_id], vram_bit);

            // Get pixel state in character
            int char_pixel = get_bit(char_data[y], x);

            // Change one single bit in the VRAM
            if (mode == MINI_OVER)
                update_bit(cpu->disp->vram[vram_id], vram_bit, char_pixel);
            else if (mode == MINI_OR)
                update_bit(cpu->disp->vram[vram_id], vram_bit, char_pixel | vram_pixel);
            else if (mode == MINI_REV)
                update_bit(cpu->disp->vram[vram_id], vram_bit, 1 - char_pixel);
            else if (mode == MINI_REVOR)
                update_bit(cpu->disp->vram[vram_id], vram_bit, (1 - char_pixel) | vram_pixel);
        }
    }
}

void draw_string_locate(cpu_t* cpu, const unsigned char* str, int max, int mode) {
    int i = 0;
    while(cpu->disp->loc_x <= 21 && cpu->disp->loc_x < max) {
        const unsigned char* c = &str[i++]; // Current character

        if (*c == 0x00) break; // Line terminator

        // Pixel start x,y on screen (128x64)
        int screen_x = (cpu->disp->loc_x - 1) * CHAR_WIDTH;
        int screen_y = (cpu->disp->loc_y - 1) * CHAR_HEIGHT;

        // Draw the character on the VRAM
        draw_character(cpu, c, screen_x, screen_y, mode);
        
        // Move the cursor to the right
        cpu->disp->loc_x++;

        // Skip the second byte of a 2-byte character
        if (is_multibyte_char(*c)) i++;
    }
}

// Display a string at the current position of the display cursor
void syscall_Print(cpu_t* cpu, uint32_t str_ptr) {
    const unsigned char* str = (const unsigned char*)get_memory_for_address(cpu, str_ptr);
    // printf("Run syscall: Print (%s)\n", str);

    draw_string_locate(cpu, str, 100, TEXT_NORMAL);
}

void syscall_PrintRev(cpu_t* cpu, uint32_t str_ptr) {
    const unsigned char* str = (const unsigned char*)get_memory_for_address(cpu, str_ptr);
    printf("Run syscall: PrintRev (%s)\n", str);

    draw_string_locate(cpu, str, 100, TEXT_REVERSE);
}

/**
 * Displays one character at the current position of the display cursor
 * @param c This is the pointer to a character string. The character string must be null-terminated.
 * @remarks After you call the PrintC function, the cursor position is moved to the right of the last character.
 */
void syscall_PrintC(cpu_t* cpu, uint32_t char_ptr) {
    const unsigned char* c = (const unsigned char*)get_memory_for_address(cpu, char_ptr);
    // printf("Run syscall: PrintC (%c)\n", *c);

    // Pixel start x,y on screen (128x64)
    int screen_x = (cpu->disp->loc_x - 1) * CHAR_WIDTH;
    int screen_y = (cpu->disp->loc_y - 1) * CHAR_HEIGHT;

    draw_character(cpu, c, screen_x, screen_y, TEXT_NORMAL);

    // Move the cursor to the right
    cpu->disp->loc_x++;
}

/**
 * Displays a character string in reversed color at the current position of the display cursor.
 * @param str This is the pointer to a character string. The character string must be null-terminated.
 * @remarks After you call the PrintRev function, the cursor position is moved to the right of the last character.
 */
void syscall_PrintRevC(cpu_t* cpu, uint32_t char_ptr) {
    const unsigned char* c = (const unsigned char*)get_memory_for_address(cpu, char_ptr);
    printf("Run syscall: PrintRevC (%s)\n", c);

    // Pixel start x,y on screen (128x64)
    int screen_x = (cpu->disp->loc_x - 1) * CHAR_WIDTH;
    int screen_y = (cpu->disp->loc_y - 1) * CHAR_HEIGHT;

    draw_character(cpu, c, screen_x, screen_y, TEXT_REVERSE);

    // Move the cursor to the right
    cpu->disp->loc_x++;
}

/**
 * Displays a character string at the specified position.
 * @param x (0 ~ 127) This is the x coordinate of the upper-left corner of the string
 * @param y (0 ~ 63) This is the y coordinate of the upper-left corner of the string
 * @param str This is the pointer to a character string. The character string must be null-terminated.
 * @param type TEXT_NORMAL or TEXT_REVERSE
 */
void syscall_PrintXY(cpu_t* cpu, int x, int y, uint32_t str_ptr, int mode) {
    const unsigned char* str = (const unsigned char*)get_memory_for_address(cpu, str_ptr);
    // printf("Run syscall: PrintXY (%d %d %s %d)\n", x, y, str, mode);

    mode = mode & 3; // Only keep the first 2 bits

    int i = 0;
    while (x < SCREEN_WIDTH) {
        const unsigned char* c = &str[i++]; // Current character

        if (*c == 0x00) break; // Line terminator

        draw_character(cpu, c, x, y, mode);

        if (is_multibyte_char(*c)) i++;

        x += CHAR_WIDTH; // Move the cursor the the right
    }
}

/**
 * Displays a small font character string at the specified position.
 * @param x (0 ~ 127) This is the x coordinate of the upper-left corner of the string
 * @param y (0 ~ 63) This is the y coordinate of the upper-left corner of the string
 * @param str This is the pointer to a character string. The character string must be null-terminated.
 * @param type MINI_OVER, MINI_OR, MINI_REV or MINI_REVOR
 */
void syscall_PrintMini(cpu_t* cpu, int x, int y, uint32_t str_ptr, int mode) {
    const unsigned char* str = (const unsigned char*)get_memory_for_address(cpu, str_ptr);
    // printf("Run syscall: PrintMini %d %d (%s, mode: %d)\n", x, y, str, mode);

    int i = 0;
    while (x < SCREEN_WIDTH) {
        unsigned char c = str[i++]; // Current character

        if (c == 0x00) break; // Line terminator

        uint8_t* char_data = cpu->disp->character_set_mini;

        // Calculate offset in the character set for 2-byte characters
        if (is_multibyte_char(c)) {
            if (c == 0xE5) char_data += CHARACTER_SET_SIZE / 4;
            else if (c == 0xE6) char_data += 2 * CHARACTER_SET_SIZE / 4;
            else if (c == 0x7F) char_data += 3 * CHARACTER_SET_SIZE / 4;
            c = str[i++];
        }
        char_data += c * 7;

        // Draw the character
        draw_character_mini(cpu, x, y, char_data, mode);
        
        // Move to the right 
        x += get_mini_char_size(char_data);
    }
}

/**
 * Displays a character string from the current position of the display cursor to the specified
 * ending x position. If the character string ends before the ending position, the space up to the specified position is
 * padded with blanks.
 * @param str This is the pointer to a character string. The character string must be null-terminated.
 * @param max (1 ~ 21) This is the ending x position.
 * @remarks After you call the PrintLine function, the cursor position is moved to the right of the last character.
 */
void syscall_PrintLine(cpu_t* cpu, uint32_t str_ptr, uint32_t max) {
    const unsigned char* str = (const unsigned char*)get_memory_for_address(cpu, str_ptr);
    // printf("[Warning] Skipped syscall: PrintLine (%s, max: %d)\n", str, max);
}

// Sets the position of the Locate cursor + display cursor
// x: [1-21], y: [1-8]
void syscall_Locate(cpu_t* cpu, int x, int y) {
    cpu->disp->loc_x = x;
    cpu->disp->loc_y = y;
    cpu->disp->cursor.col = x - 1;
    cpu->disp->cursor.row = y - 1;
    // printf("Run syscall: Locate (%d %d)\n", x, y);
}

// Sets the position of the display cursor
// column: 0..20
// row: 0..7
// The function will return 0 on failure (bad arguments), or 1 on success (arguments saved).
void syscall_Cursor_SetPosition(cpu_t* cpu, uint32_t column, uint32_t row) {
    if (column < 0 || column > 20 || row < 0 || row > 7) {
        // printf("[Warning] Cursor_SetPosition: Invalid arguments: %d %d\n", column, row);
        R0 = 0;
        return;
    }

    // printf("Run syscall: Cursor_SetPosition (%d %d)\n", column, row);

    cpu->disp->cursor.col = column;
    cpu->disp->cursor.row = row;

    R0 = 1;
}

// Returns the current cursor flash style
void syscall_Cursor_GetFlashStyle(cpu_t* cpu) {
    // printf("Run syscall: Cursor_GetFlashStyle => %d\n", cpu->disp->cursor.flash_style);
    R0 =  cpu->disp->cursor.flash_mode == 0 ? -1 : cpu->disp->cursor.flash_style;
}

// Sets the cursor flash style
// - Sets flashmode on
// - Resets to cursor-textmode (?)
void syscall_Cursor_SetFlashOn(cpu_t* cpu, uint32_t flash_style) {
    // printf("Run syscall: Cursor_SetFlashOn (%d)\n", flash_style);
    cpu->disp->cursor.flash_style = flash_style;
    cpu->disp->cursor.flash_mode = 1;
    cpu->disp->cursor.blink_state = 0;
    draw_cursor(cpu);
}

// Disables cursor flashing
void syscall_Cursor_SetFlashOff(cpu_t* cpu) {
    // printf("Run syscall: Cursor_SetFlashOff\n");
    cpu->disp->cursor.flash_mode = 0;
}

// If flashmode is 0, the Flash mode is set to 0 and cursor flashing is turned off. 
// If flashmode is nonzero, Flash mode is set to 1, and cursor flashing is turned on.
void syscall_Cursor_SetFlashMode(cpu_t* cpu, uint32_t flashmode) {
    // printf("Run syscall: Cursor_SetFlashMode (%d)\n", flashmode);
    cpu->disp->cursor.flash_mode = flashmode;
}

// Draw the cursor on the LCD screen
void draw_cursor(cpu_t* cpu) {
    cursor_t* cursor = &cpu->disp->cursor;

    if (cursor->flash_mode == 0) return;

    if(cursor->blink_state++ % 2 == 1) {
        syscall_Bdisp_PutDisp_DD(cpu);
        return;
    }
 
    int16_t id = cursor->flash_style;
    if (id == 11) id = 4;
    else if (id > 7) id -= 2;
    else if (id > 1) id -= 1;

    if (id < 0 || id > 9) {
        printf("[Warning] draw_cursor: Invalid cursor id: %d\n", id);
        id = 0;
    }

    // Pixel start x,y on screen (128x64)
    int screen_x = cursor->col * CHAR_WIDTH;
    int screen_y = cursor->row * CHAR_HEIGHT;

    for (int y = 0; y < CURSOR_ICONS_HEIGHT; y++) {
        for (int x = 0; x < CURSOR_ICONS_WIDTH; x++) {
            // Current pixel position on-screen
            int vram_x = screen_x + x;
            int vram_y = screen_y + y;
            if (vram_x < 0 || vram_x >= SCREEN_WIDTH || vram_y < 0 || vram_y >= SCREEN_HEIGHT) break;
            
            // Get pixel state in VRAM
            int vram_id = (vram_x + vram_y * SCREEN_WIDTH) / 8; // index in VRAM
            int vram_bit = 7 - vram_x % 8;                      // Current bit
            int cursor_pixel = get_bit(cpu->disp->cursor_icons[id * CURSOR_ICONS_HEIGHT + y], x);

            // Change one single bit on the LCD screen
            if (cursor_pixel) update_bit(cpu->disp->lcd[vram_id], vram_bit, cursor_pixel);
        }
    }
}