#include "headers/utils.h"

// Prints an error (same argument format as printf)
// And stops the execution of the program
void critical_error(const char* format, ...) {
    va_list args;
    va_start(args, format);
    printf("[Error] ");
    vprintf(format, args);
    va_end(args);

#ifdef USE_EMSCRIPT
    emscripten_cancel_main_loop();
#else
    exit(EXIT_FAILURE);
#endif
}

// Prints the binary representation of a number
void print_binary(uint32_t x, int n) {
    printf(" ");
    for (int i = n - 1; i >= 0; i--) {
        printf("%d", (x >> i) & 1);
    }
}

// Prints the current state of the CPU
void cpu_debug(cpu_t* cpu) {
    if (cpu->pc != 0x80010070) {
        uint16_t instruction = (mem_read(cpu, cpu->pc, 1) << 8) | mem_read(cpu, cpu->pc + 1, 1);
        printf("[%d][0x%08X] instruction: %s 0x%04X -", cpu->instruction_count, cpu->pc, get_instruction_name(instruction), instruction);
        print_binary(instruction, 16);
        printf("\n");
    }
    else printf("[%d][0x%08X] instruction: syscall\n", cpu->instruction_count, cpu->pc);
}

// Prints the current state of the VRAM
void vram_debug(cpu_t* cpu) {
    for (int y = 0; y < 64; y++) {
        for (int x = 0; x < 16; x++) {
            for (int b = 7; b >= 0; b--) {
                int id = x + y * 16;
                if (id < 0 || id >= VRAM_SIZE) critical_error("error %d", id);
                int bit = ((cpu->disp->vram[id] >> b) & 1) || y == 0 || y == 63 || (x == 0 && b == 7) || (x == 15 && b == 0);
                
                printf(bit ? "_" : "X");
            }
        }
        printf("\n");
    }
    printf("\n");
}

// Debug function called from javascript
char* get_next_instruction(cpu_t* cpu) {
    if (cpu->pc == cpu->pr) return "<EOF>";
    if (cpu->pc == SYSCALL_ADDRESS) return "<syscall>";

    uint16_t instruction = (mem_read(cpu, cpu->pc, 1) << 8) | mem_read(cpu, cpu->pc + 1, 1);
    char* instruction_name = calloc(64, sizeof(char));
    
    strcpy(instruction_name, get_instruction_name(instruction));
    strcat(instruction_name, " 0x");
    char* instruction_hex = calloc(5, sizeof(char));
    sprintf(instruction_hex, "%04X", instruction);
    strcat(instruction_name, instruction_hex);
    free(instruction_hex);

    return instruction_name;
}